#!/usr/bin/env python3

# ¡Py3!

import os
import shutil
import string
import sys

INITS = string.digits + string.ascii_uppercase

AVOID_BECAUSE_TOO_BIG = [
    '/home/facundo/Música/Discos/Battlestar Galactica',
    '/home/facundo/Música/Les Luthiers',
    '/home/facundo/Música/Infantiles/'
    'André Popp/Piccolo, Saxo y compañía/Música sin texto',
]

DONT_FLATTEN = {'Recitales', 'Fiesta'}


def _sanitize(path):
    """Fix a path to make it filesystems friendly."""
    path = path.replace(":", " ")
    path = path.replace("¿", "")
    path = path.replace("?", "")
    return path


def exists(path):
    """An os.path.exists() that retries on sanitized path."""
    if os.path.exists(path):
        return True
    return False


def get_content(path):
    """Get the content of a file, retrying on sanitized path."""
    fh = open(path, "rb")

    try:
        content = fh.read()
    except OSError:
        print("Problem reading file {!r}".format(path))
        raise
    fh.close()
    return content


def main(sourcedir, destdir, restrictedsize, dryrun, flatten,
         name_range, reverse=False, contentcheck=False, extradir=False):
    """Main entry point."""
    # name mangling in the code depends of these being absolute
    sourcedir = os.path.abspath(sourcedir)
    destdir = os.path.abspath(destdir)

    for dirpath, dirnames, filenames in os.walk(sourcedir, followlinks=True):
        # dirty hack to avoid some directories instead of converting them
        # on the fly
        if restrictedsize:
            should_continue = False
            for restrictedpath in AVOID_BECAUSE_TOO_BIG:
                if dirpath.startswith(restrictedpath):
                    should_continue = True
                    break
            if should_continue:
                continue

        dirnames.sort()
        filenames.sort()
        pushing = dirpath[len(sourcedir) + 1:]
        if name_range is not None:
            if pushing:
                # check the base dir, the one we're pushing
                if pushing[0] not in name_range:
                    continue
            else:
                # we're in the base dir, check dir and filenames
                dirnames = [d for d in dirnames if d[0] in name_range]
                filenames = [f for f in filenames if f[0] in name_range]
        if extradir and pushing:
            converted = os.path.join(destdir, pushing[0].upper(), pushing)
        else:
            converted = os.path.join(destdir, pushing)

        # check dirs
        if not flatten:
            for dname in dirnames:
                if extradir and os.path.normpath(converted) == os.path.normpath(destdir):
                    should = os.path.join(converted, dname[0].upper(), dname)
                else:
                    should = os.path.join(converted, dname)
                if os.path.exists(should):
                    if not os.path.isdir(should):
                        print("WARNING: should be a directory and it's not",
                              repr(should))
                else:
                    if reverse:
                        fullpath = os.path.join(dirpath, dname)
                        print("  # Reverse: Dir missing:\n  rmdir", repr(fullpath))
                        continue

                    if dryrun:
                        print("# directory missing, creating")
                        print("  mkdir", repr(should))
                    else:
                        print("Directory missing, creating", repr(should))
                        os.makedirs(should)

        # check files
        for fname in filenames:
            srcfile = os.path.join(dirpath, fname)
            if flatten and pushing:
                if os.path.basename(dirpath) in DONT_FLATTEN:
                    dstfile = "{}/{}".format(converted, fname)
                else:
                    dstfile = "{} - {}".format(converted, fname)
            else:
                dstfile = os.path.join(converted, fname)
            if exists(dstfile):
                if os.path.isdir(dstfile):
                    print("WARNING: should not be a directory and it is", repr(dstfile))
                    continue

                if reverse:
                    continue

                # file is there, and not a directory, assure have same size
                srcsize = os.stat(srcfile).st_size
                dstsize = os.stat(dstfile).st_size
                if srcsize == dstsize:
                    if not contentcheck:
                        # size comparison is enough
                        continue
                    srccontent = get_content(srcfile)
                    dstcontent = get_content(dstfile)
                    if srccontent == dstcontent:
                        # even the content is the same!
                        continue
                    else:
                        reason = "File content is different, copying again"
                else:
                    reason = "File size is different, copying again"
            else:
                if reverse:
                    print("  # Reverse: File missing:\n  rm", repr(srcfile))
                    continue
                reason = "File missing, copying"

            if dryrun:
                print("  #", reason)
                size = os.stat(srcfile).st_size / 1024 ** 2
                print("  cp -p {!r} {!r}  # {:.1f} MB".format(srcfile, dstfile, size))
            else:
                print(reason, repr(dstfile))
                try:
                    shutil.copy2(srcfile, dstfile)
                except OSError as err:
                    if err.errno == 95:
                        # operation not supported (changing stat info in some dest filesystems)
                        pass
                    else:
                        raise


if __name__ == '__main__':
    argv = sys.argv[1:]
    try:
        argv.remove('--restrictedsize')
    except ValueError:
        restricted = False
    else:
        restricted = True
    try:
        argv.remove('--noreverse')
    except ValueError:
        reverse = True
    else:
        reverse = False
    try:
        argv.remove('--contentcheck')
    except ValueError:
        contentcheck = False
    else:
        contentcheck = True
    try:
        argv.remove('--dry')
    except ValueError:
        dryrun = False
    else:
        dryrun = True
    try:
        argv.remove('--flatten')
    except ValueError:
        flatten = False
    else:
        flatten = True
    try:
        argv.remove('--extra-dir')
    except ValueError:
        extradir = False
    else:
        extradir = True

    try:
        pos = argv.index("--range")
    except ValueError:
        name_range = None
    else:
        rf = argv[pos + 1].upper()
        rt = argv[pos + 2].upper()
        pf = INITS.index(rf)
        pt = INITS.index(rt)
        name_range = INITS[pf:pt + 1]
        del argv[pos:pos + 3]

    if len(argv) != 2:
        print("Usage: musynch.py")
        print("             [--dry] [--restrictedsize] [--range X Y] [--extra-dir] [--noreverse]")
        print("             sourcedir remotedir")
        exit()

    main(argv[0], argv[1], restricted, dryrun, flatten, name_range,
         contentcheck=contentcheck, extradir=extradir)
    if reverse:
        main(argv[1], argv[0], restricted, dryrun, flatten, name_range,
             reverse=True, extradir=extradir)
